; COMPUTATION OF PI, NATURAL LOGARITH OF 2 (ln 2) AND THE NUMBER e BY 
; DARIO A. ALPERN USING DPMI (DOS PROTECTED MODE INTERFACE)
;
; Formulas used:
; =============
;
; Computation of PI:
;
;                   1            1            1             13753  
; pi = 4  6 arc tg  + arc tg  + arc tg  + arc tg  
;                   8           32           128          63664921 
;
;
; Computation of ln 2:
;
;                    1             1             1529
; ln 2 =  5 arg tgh  + arg tgh  + arg tgh 
;                    8            16            670751
;
;
; The arc tangent is computed using Taylor formula:
;
;                  3     5     7     9
;                 x     x     x     x
; arc tg x = x -  +  -  +  - ...
;                 3     5     7     9
;
; The argument of the hyperbolic tangent is also computed using
; Taylor formula:
;
;                   3     5     7     9
;                  x     x     x     x
; arg tgh x = x +  +  +  +  + ...
;                  3     5     7     9
;
;
; These formulas can be found by noticing that:
;
;                                                           x + y
;   pi = 4 arc tg 1     and   arc tg x + arc tg y = arc tg  ;
;                                                           1 - xy
;
;                 3                                             x + y
; ln 2 = arg tgh     and   arg tgh x + arg tgh y = arg tgh  .
;                 5                                            1 + xy
;
;----------------------------------------------------------------------------
;
; Three buffers that hold numbers are used for the computation of PI or ln 2.
; The first one contains the partial sum, the second holds the powers of
; the last arc tg or arg tgh, and the third holds the number that will add
; or subtract the number in the first buffer.
;
;----------------------------------------------------------------------------
;
; Definitions of constants and initialized variables in program.
;
.386
eloop    macro destination        ;Macro that makes a loop in ECX.
         dec ecx                  ;;This is faster than the LOOP instruction.
         jne destination
         endm

struc_real_mode struc         ;Structure to pass parameters to real mode.
reg_EDI  dd ?                 ;Register EDI image.
reg_ESI  dd ?                 ;Register ESI image.
reg_EBP  dd ?                 ;Register EBP image.
reserved dd ?
reg_EBX  dd ?                 ;Register EBX image.
reg_EDX  dd ?                 ;Register EDX image.
reg_ECX  dd ?                 ;Register ECX image.
reg_EAX  dd ?                 ;Register EAX image.
reg_Flags  dw ?               ;Flags image.
reg_ES   dw ?                 ;Register ES in real mode image.
reg_DS   dw ?                 ;Register DS in real mode image.
reg_FS   dw ?                 ;Register FS in real mode image.
reg_GS   dw ?                 ;Register GS in real mode image.
reg_IP   dw ?                 ;Register IP image.
reg_CS   dw ?                 ;Register CS in real mode image.
reg_SP   dw ?                 ;Register SP image.
reg_SS   dw ?                 ;Register SS in real mode image.
struc_real_mode ends

HANDLE_STDOUT      equ 1      ;Standard Output Handle (DOS).
HANDLE_STDERR      equ 2      ;Standard Error Handle (DOS).

ASCII_CR           equ 13     ;Move cursor to left column (Carriage Return).
ASCII_LF           equ 10     ;Move cursor to next line (Line Feed).
port6845           equ 3D4h   ;Video controller port.

PIDPMIE segment use16
        assume cs:PIDPMIE,ds:PIDPMIE
        org 100h            ;Start offset for .COM programs.
begin:  jmp start           ;Jump to the beginning of program.

divisor         dd 1        ;Divisor of each term. It takes the values
                            ;1, 3, 5,... (see the formulas).

digits          dd 0        ;Number of digits of the output.

PMode_entry     label dword ;Entry point address to protected mode.
PMode_entry_offs dw 0       ;Offset.
PMode_entry_segm dw 0       ;Segment.

NbrDWords       dd 0        ;Number of DWords (32 bits) equivalent to the
                            ;number of requested digits.

LastDWord       dd    0
FirstDWord      dd    0     ;Most significant DWord not zero.
                            ;The speed is doubled if the DWords in zero
                            ;are not processed.

Numerator       dd 13753*13753 ;Numerator (32 bits) of the argument of last
                            ;arc tg/arg tgh. It is initialized for the case
                            ;of PI. In the other case this value is
                            ;overwritten with 1529*1529 at the beginning 
                            ;of the program. See formulas above.

Denominator     dd 63664921 ;Denominator (32 bits) of the argument of last
                            ;arc tg/arg tgh. It is initialized at the
                            ;beginning of the program to 63664921 (for PI) 
                            ;or 670751 (for ln 2). Ver formulas arriba.

; Multipliers used to find the number of DWords W needed to
; compute n digits in base b: W = 2^32*log(b)/log(2^32)
Multipliers label dword
                dd 134217728    ; Base 2
                dd 212730066    ; Base 3
                dd 268435456    ; Base 4
                dd 311643913    ; Base 5
                dd 346947794    ; Base 6
                dd 376796799    ; Base 7
                dd 402653184    ; Base 8
                dd 425460132    ; Base 9
                dd 445861641    ; Base 10
                dd 464317052    ; Base 11
                dd 481165522    ; Base 12
                dd 496664612    ; Base 13
                dd 511014527    ; Base 14
                dd 524373979    ; Base 15
                dd 536870912    ; Base 16
                dd 548609976    ; Base 17
                dd 559677860    ; Base 18
                dd 570147180    ; Base 19
                dd 580079369    ; Base 20
                dd 589526865    ; Base 21
                dd 598534780    ; Base 22
                dd 607142208    ; Base 23
                dd 615383250    ; Base 24
                dd 623287827    ; Base 25
                dd 630882340    ; Base 26
                dd 638190197    ; Base 27
                dd 645232255    ; Base 28
                dd 652027172    ; Base 29
                dd 658591707    ; Base 30
                dd 664940973    ; Base 31
                dd 671088640    ; Base 32
                dd 677047118    ; Base 33
                dd 682827704    ; Base 34
                dd 688440713    ; Base 35
                dd 693895588    ; Base 36

MaxPower        dd 0        ;Maximum power in the requested base that fits 
                            ;in a doubleword.
                            ;The value is: Base^(int(log (2^32)/log Base))
LogMaxPower     dd 0        ;Exponent for the power shown above.
                            ;Number of digits per DWord.
                            ;The value is: int(log (2^32)/log Base)) *
                            ;* 2^32*log Base/log(2^32)
SumLogPower     dd 0        ;Sum of LogMaxPower in order to determine the
                            ;number of digits during the conversion.
Exponent        dd 0        ;Indicates the number of digits that are
                            ;converted in each step.
MemoryHandleLow dw 0        ;Memory handle given by the DPMI server.
MemoryHandleHigh dw 0

PointerSubrPowersOf2 dw BinaryPowersForPI
                            ;Pointer to subroutine that computes the binary
                            ;powers. It is initialized for the calculation 
                            ;of PI. To compute ln 2 this value is overwritten
                            ;with the corresponding pointer.

handle          dw 1        ;Handle for the Standard Output (the screen by
                            ;default). It is overwritten with the file
                            ;handle (if it is specified).

CommandLineSwitches db 0    ;It contains bitfields for each command line
                            ;switch. The masks are the following:
LN2                equ 1
E                  equ 2
Format             equ 4
ShowNumberOfDigits equ 8


bufferdos       db "3.",60 dup (0) ;Buffer for conversion from binary to
                                   ;another base.

FormattedNumberBuffer db "3.",75 dup (" "),ASCII_CR,ASCII_LF
                            ;One-line buffer used to obtain the formatted
                            ;output.

PositionDigitInBuffer dw FormattedNumberBuffer + 2
                            ;Pointer to the previous buffer.

FormattingInformation dw 0    ;MSB = Group number (0-9).
                              ;LSB = Digit inside group (1-5).

DigitsShown          dd 0

Selector             dw 0

Initials       db "dhms"
Days           db 0
Hours          db 0
Minutes        db 0
Seconds        db 0
FracSecond     dd 0

OldInt8Handler label dword
OldInt8Offset  dw 0
OldInt8Segment dw 0

ShowInInt8     db 1
PointerBase    dw 0
Base           db 10


CursorVideo    label dword
CursorVideoOffset  dw 0
CursorVideoSegment dw 0B800h

CopyRight db "Computation of PI, ln 2 and e. Written by Dario Alejandro Alpern."
        db ASCII_CR,ASCII_LF
LengthCopyRight equ $ - CopyRight
Text    db ASCII_CR,ASCII_LF
        db "Command line: PIDPMIE number_of_figures [/F] [/LN2] [/E] "
        db "[/B:base]",ASCII_CR,ASCII_LF," [dest_file]",ASCII_CR,ASCII_LF
        db "   /F: Give formatted output suitable for text editors"
        db ASCII_CR,ASCII_LF
        db "   /LN2: Shows the value of the natural logarithm of 2 instead of pi"
        db ASCII_CR,ASCII_LF
        db "   /E: Shows the value of the number e instead of pi"
        db ASCII_CR,ASCII_LF
        db "   /B: Shows the output in any base 2 to 36."
        db ASCII_CR,ASCII_LF
        db "   dest_file: File to which the number will be written"
        db ASCII_CR,ASCII_LF
        db "              If no file is given the output will go to the screen (STDOUT)."
        db ASCII_CR,ASCII_LF
        db ASCII_CR,ASCII_LF
LengthText equ $ - Text
NoDPMIServer db "There is no DPMI server.",ASCII_CR,ASCII_LF
LengthNoDPMIServer equ $ - NoDPMIServer
DPMI16BitsServer db "The DPMI server does not support 32-bit applications."
               db ASCII_CR,ASCII_LF
LengthDPMI16BitsServer equ $ - DPMI16BitsServer
OutOfMemoryForServer db "There is not enough memory for DPMI server."
               db ASCII_CR,ASCII_LF
LengthOutOfMemoryForServer equ $ - OutOfMemoryForServer
CannotSwitchToProtMode db "Cannot switch to protected mode."
               db ASCII_CR,ASCII_LF
LengthCannotSwitchToProtMode equ $ - CannotSwitchToProtMode
CannotReserveSelector db "Cannot reserve selector."
               db ASCII_CR,ASCII_LF
LengthCannotReserveSelector equ $ - CannotReserveSelector
OutOfMemory db ASCII_CR,ASCII_LF
        db "There is not enough memory. Try reducing the number of figures. "
        db ASCII_CR,ASCII_LF
LengthOutOfMemory equ $ - OutOfMemory
TextSecondLine db "Digits found: "
LengthTextSecondLine equ $ - TextSecondLine
TotalTime db ASCII_CR,ASCII_LF,"Total time: "
LengthTotalTime equ $ - TotalTime
RealMode struc_real_mode <0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0>
;=============================================================================
; Show Copyright
;=============================================================================
start:  mov dx,offset CopyRight     ;Show copyright on screen.
        mov cx,LengthCopyRight
        mov bx,HANDLE_STDERR        ;Handle STDERR.
        mov ah,40h
        int 21h
;==========================================================;
; Check if a DPMI server exists and go to protected mode   ;
;==========================================================;
CheckDPMI:
        mov RealMode.reg_DS,ds
        pushf
        pop ax
        and ax,0FFFh
        mov RealMode.reg_Flags,ax
        mov ax,1687h              ;Check if a DPMI server is installed
        int 2Fh                   ;in memory.
        test ax,ax
        jz short InitializeDPMI   ;Jump if there is a DPMI server.
        mov dx,offset NoDPMIServer
        mov cx,LengthNoDPMIServer ;Tell the user that there is no DPMI server.
error_in_real_mode:
        mov bx,HANDLE_STDERR    ;Handle STDERR.
        mov ah,40h
        int 21h
        mov ax,4c01h            ;End program with return code 1.
        int 21h
InitializeDPMI:
        mov PMode_entry_segm,es ;Preserve protected mode entry point 
        mov PMode_entry_offs,di ;of DPMI server.
        test bl,1               ;Does the server support 32-bit applications?
        mov dx,offset DPMI16BitsServer
        mov cx,LengthDPMI16BitsServer
        jz short error_in_real_mode  ;Jump if not.
        test si,si              ;Does the server need local memory?
        jz short go_to_protected_mode ;Jump if not.
        push ds
        pop es                  ;ES = Segment of program .COM
        push si                 ;Preserve the number of bytes needed by
                                ;the DPMI server.
        mov ah,4Ah              ;Shrink the DOS memory block that
        mov bx,offset EndCode[15] ;includes the program .COM
        shr bx,4
        int 21h
        pop bx
        mov ah,48h
        int 21h                 ;Reserve memory for the DPMI server.
        mov cx,LengthOutOfMemoryForServer
        mov dx,offset OutOfMemoryForServer
        jc short error_in_real_mode  ;Jump if memory could not be reserved.
        mov es,ax               ;Point to data area of the DPMI server.
go_to_protected_mode:
        mov ax,1                ;Indicate that this application is 32-bit.
        call PMode_entry        ;Switch to protected mode.
        mov cx,LengthCannotSwitchToProtMode
        mov dx,offset CannotSwitchToProtMode
        jc short error_in_real_mode   ;Jump if the switch could not be done.
        push ds
        pop es                  ;Volver a poner ES = DS.
        xor ax,ax
        mov cx,1                ;Reserve a descriptor.
        int 31h
        mov cx,LengthCannotReserveSelector
        mov dx,offset CannotReserveSelector
        jc short error_in_prot_mode   ;Jump if the descriptor could not be
                                      ;reserved.
        mov Selector,ax         ;Preserve the corresponding selector.
        mov bx,ax
        mov ax,0009h
        mov cx,40F2h
        int 31h                 ;Set the Access Rights byte.
HallarCantDecimales:
        mov bx,80h              ;Point to previous to first byte of command
                                ;line.
        mov esi,10              ;ESI = Multiplier.
        xor eax,eax             ;Reset number of digits requested by user.
        xor ecx,ecx             ;Initialize digit to zero.
SearchForDigit:
        inc bx                  ;Point to next byte in command line.
        mov al,[bx]             ;Get character from command line.
        cmp al,ASCII_CR         ;End of line?
        je short syntax_error   ;Jump to error if it is.
        cmp al," "              ;Space or unprintable character?
        jbe SearchForDigit      ;If so, skip that character.
        sub al,"0"              ;Is the character numeric (range '0' to '9')?
        cmp al,9
        jbe short CheckIfNewDigit ;Jump if it is numeric.
                                ;Otherwise show command line format since
                                ;the command line must start with a digit.
syntax_error:
        mov dx,offset Text      ;Show command line format.
        mov cx,LengthText
error_in_prot_mode:
        mov bx,HANDLE_STDERR    ;Handle STDERR.
        mov ah,40h
        call dos                ;Call DOS from protected mode.
        mov ax,4c01h            ;End of program with retrun code 1.
        int 21h
NewDigit:
        mul esi                 ;Multiply number of requested digits by 10.
        jc NotEnoughMemory      ;Jump if the number is too big.
        add eax,ecx             ;Add digit found in command line.
        jc NotEnoughMemory      ;Jump if the number is too big.
CheckIfNewDigit:
        inc bx                  ;Point to next character in command line.
        mov cl,[bx]             ;Get character from command line.
        sub cl,"0"              ;Check if it is numeric.
        cmp cl,9
        jbe NewDigit            ;Jump if there is a new digit.
        mov digits,eax          ;Store number of digits in variable.
;
; Check in the command line for the presence of the switches /F, /LN2 or /E.
; According to the present switches, set the correponding bits in the
; variable CommandLineSwitches.
;
        call switch             ;Set flags according to the switches found.
        xor ecx,ecx
        mov cl,Base
        mov di,cx               ;Get base.
        shl di,2
        add di,offset Multipliers[-8] ;Point to logarithm.
        mov PointerBase,di
        mov si,-1               ;Initialize exponent.
        mov eax,1               ;Initialize power.
LoopFindPower:
        inc si                  ;Actualize exponent.
        mul ecx
        jnc LoopFindPower
        mov word ptr Exponent,si
        div ecx
        mov MaxPower,eax        ;Store maximum power.
        mov eax,Exponent
        mul dword ptr [di]
        mov LogMaxPower,eax
HallarNbrDWords:
        mov eax,digits          ;EAX = Number of digits to compute.
        mov bx,PointerBase      ;Get pointer to multiplier corresponding
                                ;to the requested base.
        mul dword ptr [bx]      ;EDX = Number of needed DWords.
        add edx,3               ;Reserve 3 more DWords for rounding.
        mov NbrDWords,edx       ;Store number of needed DWords.
        shl edx,2               ;Convert to number of bytes.
        lea edx,[edx * 2 + edx] ;Reserve space for 3 buffers.
        or dx,0FFFh             ;Compute the limit for 4KB page complete.
                                ;The limit of a segment is its length - 1.
        shld ecx,edx,16         ;CX = High word of limit.
        push cx                 ;Preserve high word of limit.
        push dx                 ;Preserve low word of limit.
        mov bx,Selector
        mov ax,0008h
        int 31h                 ;Store segment limit in descriptor.
        pop cx                  ;Restore low word of limit.
        pop bx                  ;Restore high word of limit.
        add cx,1                
        adc bx,0                ;BX:CX = Number of bytes of segment.
        mov ax,0501h            ;Reserve this number of bytes of linear memory.
        int 31h                 ;BX:CX = Linear address of memory block.
                                ;SI:DI = Memory handle.
        jnc short EnoughMemory ;Jump if there were no errors.
NotEnoughMemory:
        mov dx,offset OutOfMemory ;Indicate error condition on screen.
        mov cx,LengthOutOfMemory
        mov ah,40h
        call dos                ;Call DOS from protected mode.
        mov ax,4C02h            ;End program with error code 2.
        int 21h
EnoughMemory:
        mov MemoryHandleLow,di  ;Preserve memory handle.
        mov MemoryHandleHigh,si
        mov dx,cx               ;DX = Lower word of segment base.
        mov cx,bx               ;CX = Higher word of segment base.
        mov ax,0007h
        mov bx,Selector
        int 31h                 ;Store segment base in descriptor.
InicializarPrimerTermino:
        test CommandLineSwitches,E  ;Check if the number e is to be computed.
        jz ItsNotE
;------------------------------------------------------------------------------
; Computation of number e:
;
        mov bufferdos,"2"       ;Integer part of e.
        call SetInterruptVector
        mov ebx,NbrDWords
        mov ecx,ebx             ;Obtain number of DWords of number.
        shl ecx,1               ;Obtain number of DWords to set to zero.
        xor edi,edi             ;Point to beginning of number located
                                ;in memory given by DPMI server.
        xor eax,eax             ;Set both numbers to zero.
        rep stos dword ptr es:[edi]
        inc divisor                    ;Divisor = 2.
        mov byte ptr es:[ebx*4+3],80h  ;Set term to 0.5
        mov byte ptr es:[3],80h        ;Set sum to 2.5
MainLoopE:
        mov ecx,NbrDWords              ;ECX = Number of DWords of number.
        inc divisor                    ;Next divisor.
        mov edi,ecx                    ;EDI = Number of DWords of number.
        shl edi,2                      ;EDI = Pointer to term.
        mov eax,FirstDWord             ;EAX = Index of the most significant
                                       ;      DWord not zero.
        sub ecx,eax                    ;ECX = Number of DWords to process.
        jbe BinaryToDecimal            ;Jump if it is <= 0.
        lea edi,[edi+eax*4]            ;EDI = Pointer to the most significant
                                       ;      DWord not zero.
        mov esi,divisor                ;ESI = Divisor in the "e" formula.
        xor edx,edx                    ;EDX = Division reminder.
        mov ebx,edi                    ;EBX = Pointer to most significant
                                       ;      DWord.
        cld                            ;Process in ascending addresses
                                       ;(to perform a division the most
                                       ;significant DWords should be taken
                                       ;first).
        push ecx                       ;Preserve number of Dwords to
                                       ;process.
        shr ecx,1
        jnc short DivVariableE         ;Jump if number of DWords is even.
        mov eax,es:[edi]               ;Obtain DWord from dividend.
        div esi                        ;Make first division.
        stos dword ptr es:[edi]        ;Store quotient.
DivVariableE:
        jecxz EndDivVariableE          ;Jump if division is finished.
LoopDivVariableE:
        mov eax,es:[edi]               ;Obtain DWord from dividend.
        div esi                        ;Make first division in loop.
        stos dword ptr es:[edi]        ;Store quotient.
        mov eax,es:[edi]               ;Obtener el DWord del dividendo.
        div esi                        ;Make second division in loop.
        stos dword ptr es:[edi]        ;Store quotient.
        eloop LoopDivVariableE         ;Close loop.
EndDivVariableE:
        cmp dword ptr es:[ebx],1       ;Most significant DWord is zero?
        adc FirstDWord,ecx             ;MSD is now the next if it is zero.
        std                            ;Process in descending addresses
                                       ;(to perform a sum the least
                                       ;significant DWords should be taken
                                       ;first).
        lea esi,[edi-4]                ;Point to LSD of term.
        mov eax,NbrDWords              ;EAX = Number of DWords of number.
        shl eax,2                      ;EAX = Number of bytes of number.
        mov edi,esi
        sub edi,eax                    ;EDI = Pointer to LSD of sum.
        pop ecx                        ;Restore number of Dwords to process.
        shr ecx,1
        jnc short SumE                 ;Jump if number of DWords is even.
        lods dword ptr es:[esi]        ;Get DWord from term.
        adc eax,dword ptr es:[edi]     ;Add it to the sum.
        stos dword ptr es:[edi]        ;Store it in the sum.
SumE:
        jecxz EndSumE                  ;Jump if sum is finished.
LoopSumE:
        lods dword ptr es:[esi]        ;Get DWord from term.
        adc eax,dword ptr es:[edi]     ;Sumarselo a la suma.
        stos dword ptr es:[edi]        ;Add it to the sum.
        lods dword ptr es:[esi]        ;Store it in the sum.
        adc eax,dword ptr es:[edi]     ;Add it to the sum.
        stos dword ptr es:[edi]        ;Store it in the sum.
        eloop LoopSumE                 ;Close loop.
EndSumE:
        jnc MainLoopE                  ;Jump if sum is finished.
LoopSumCarryE:
        inc dword ptr es:[edi]         ;Increment next DWord more
                                       ;significant.
        lea edi,[edi-4]                ;Point to next DWord more
                                       ;significant.
        jnz MainLoopE                  ;Jump if sum is finished.
        jmp LoopSumCarryE              ;Continue adding.
;------------------------------------------------------------------------------
ItsNotE:
        mov eax,13753*4           ;First numerator for PI.
        mov Numerator,13753*13753 ;Numerator for PI.
        mov Denominator,63664921  ;Denominator for PI.
        test CommandLineSwitches,LN2    ;Check what number is to be computed.
        jz short EndInitializationConst ;Jump if it is the number PI.
        mov PointerSubrPowersOf2,offset BinaryPowersForLN2
                                 ;Subroutine to be called when the powers
                                 ;of 2 will be computed.
        mov eax,1529             ;First numerator for LN 2.
        mov Numerator,1529*1529  ;Numerator for LN 2.
        mov Denominator,670751   ;Denominator for LN 2.
        mov bufferdos,"0"        ;Indicate that the number starts with zero.
EndInitializationConst:
        push eax                 ;Preserve the first numerator.
        call SetInterruptVector
        pop edx                  ;Get the first numerator.
        mov ecx,NbrDWords        ;Obtain number of DWords of number.
        lea esi,[ecx*4]          ;Convert to bytes.
        xor edi,edi              ;Point to beginning of number located
                                 ;in memory given by DPMI server.
        mov es,Selector          ;Register ES points to that memory.
        mov ebx,Denominator      ;Denominator of first term.
        cld                      ;Process in ascending addresses.
LoopCalcFirstTerm:
        xor eax,eax              ;The decimal part of denominator is zero
                                 ;in the first term.
        div ebx                  ;Generate DWord of quotient.
        mov es:[edi+esi],eax     ;Store it in the second number.
        stos dword ptr es:[edi]  ;Store it in the first number.
        eloop LoopCalcFirstTerm  ;Close loop.
        mov al,256*4/32+256*4/128     ;Add powers of 2 for Pi.
        test CommandLineSwitches,LN2  ;Is ln 2 to be calculated?
        jz short PutFirstPowerOf2     ;Jump if it is not.
        mov al,256*5/8+256*1/16       ;Add powers of 2 for LN 2.
PutFirstPowerOf2:
        mov es:[3],al                 ;Store it in the first number.
        mov es:[esi+3],al             ;Store it in the second number.
MainLoop:
        call NewTerm                  ;Compute new term.
        test CommandLineSwitches,LN2  ;Is LN 2 to be computed?
        jnz short CicloSumaDelTermino ;Jump if so, because all terms in
                                      ;the Taylor formula for arg tgh are
                                      ;positive.
LoopSubtractFromTerm:
        mov eax,es:[edi]           ;Obtain DWord of the sum.
        sbb eax,es:[esi]           ;Subtract Dword from subtrahend.
        stos dword ptr es:[edi]    ;Store this DWord in the sum.
        lea esi,[esi-4]            ;Point to next DWord.
        eloop LoopSubtractFromTerm ;Close loop.
        jnc short TermForSum       ;Jump if the subtraction is finished.
ContinueSubtract:
        sbb dword ptr es:[edi],ecx ;Actualize DWord.
        lea edi,[edi-4]            ;Actualize pointer.
        jc ContinueSubtract        ;Jump if subtraction is not finished.
TermForSum:
        call NewTerm               ;Compute new term.
CicloSumaDelTermino:
        lods dword ptr es:[esi]    ;Obtain DWord of term.
        adc eax,es:[edi]           ;Add it to DWord of the sum.
        stos dword ptr es:[edi]    ;Store DWord of sum.
        eloop CicloSumaDelTermino  ;Close loop.
        jnc MainLoop               ;Jump if addition is not finished.
ContinueAdding:
        adc dword ptr es:[edi],ecx ;Actualize DWord.
        lea edi,[edi-4]            ;Actualize pointer.
        jc ContinueAdding          ;Jump if addition is not finished.
        jmp MainLoop               ;Close main loop.
;--------------------------------------------------------------------------
; Subroutine BinaryPowersForLN2: Adds or subtracts binary powers of 2 for
; the calculation of the natural logarithm of 2.
; It works by "xoring" with the appropiate mask.
;
BinaryPowersForLN2:
        mov edi,divisor
        lea edi,[edi*2+edi-3]
        mov cx,di
        mov eax,80000000h
        ror eax,cl
        shr edi,3
        and di,0FFFCh
        mov ecx,NbrDWords
        xor es:[edi+ecx*4],eax
        mov eax,3
        mul divisor
        call ChangeBit
        mov eax,4
        mul divisor
        jmp short ChangeBit
;----------------------------------------------------------------------------
; Subroutine BinaryPowersForPI: Adds or subtracts binary powers of 2 for
; the calculation of Pi.
; It works by "xoring" with the appropiate mask.
;
BinaryPowersForPI:
;
; This subroutine should calculate 24 / 8 ^ (div), where "div" is the divisor 
; (3, 5, 7,..) of the term. To obtain the position of the bits to perform
; the "xor" operation we have to note the following:
;
;          24 / 8 ^ (div) = 3 / 8 ^ (div - 1)
;                         = 3 / 2 ^ (3 * div - 3)
;                         = 2 ^ -(3 * div - 4) + 2 ^ -(3 * div - 3)
;
; Therefore the bits to change are the following negative powers of two:
; 3 * div - 4 y 3 * div - 3.
;
; Notice that the bits are numbered in the following way:
; bit 0 -> 2^(-1), bit 1 -> 2^(-2), etc., so the bit numbers to change
; are 3 * div - 5 y 3 * div - 4.
;
        mov edi,divisor             ;EDI <- div
        lea edi,[edi*2+edi]         ;EDI <- 3 * div
        sub edi,5                   ;EDI <- 3 * div - 5
        jb short OtherPowers        ;There are carry only when the divisor is
                                    ;1, but in this case 24 / 8^div is a
                                    ;integer number, so nothing is to be done.
        mov cx,di                   ;Bits 4-0 of CL: Bit position.
        mov eax,0C0000000h          ;Inicialize for rotation.
        ror eax,cl                  ;Perform rotation.
        shr edi,3                   ;Convert bit number to byte number.
        and di,0FFFCh               ;DWord align.
        mov ecx,NbrDWords           ;ECX = Number of DWords of number.
        xor es:[edi+ecx*4],eax      ;Perform "xor" in order to add or
                                    ;subtract binary power.
OtherPowers:
        mov eax,5                   ;Add or subtract power of (1/32).
        call Power
        mov eax,7                   ;Add or subtract power of (1/128).
Power:
        mul divisor
        sub eax,2
ChangeBit:
        mov cl,al
        inc edx
        ror edx,cl
        shr eax,5
        mov ecx,NbrDWords
        cmp eax,ecx
        lea eax,[eax*4]
        jae short FinSubrutina
        xor es:[eax+ecx*4],edx
FinSubrutina:
        ret
;-----------------------------------------------------------------------------
NewTerm:
        call PointerSubrPowersOf2      ;Subtract powers of 2 from
                                       ;previous term.
        mov edi,ecx                    ;EDI = Number of DWords of number.
        shl edi,2                      ;EDI = Pointer to term.
        mov eax,FirstDWord             ;EAX = Index of the most significant
                                       ;      DWord not zero.
        sub ecx,eax                    ;ECX = Number of DWords to process.
        jbe short NumberOfDWordsOfTermOk ;Jump if it is <= 0.
        lea edi,[edi+eax*4]            ;EDI = Pointer to the most significant
                                       ;      DWord not zero.
        mov esi,Denominator            ;ESI = Denominator of arc tg/arg tgh.
        xor edx,edx                    ;EDX = Remainder of the first division
                                       ;by the denominator.
        xor ebx,ebx                    ;EBX = Remainder of the second division
                                       ;by the denominator.
        cld                            ;Process in ascending addresses
                                       ;(to perform a division the most
                                       ;significant DWords should be taken
                                       ;first).
        push ecx                       ;Preserve number of DWords to process.
LoopDivByDenominatorSquared:
        mov eax,es:[edi]               ;Obtain DWord from dividend.
        div esi                        ;Perform first division.
        xchg ebx,edx                   ;EDX = Remainder of the previous
                                       ;second division.
                                       ;EBX = Remainder of the actual
                                       ;first division.
        div esi                        ;Perform second division.
        xchg ebx,edx                   ;EDX = Remainder of the actual
                                       ;first division.
                                       ;EBX = Remainder of the actual
                                       ;second division.
        stos dword ptr es:[edi]        ;Store quotient of second division.
        eloop LoopDivByDenominatorSquared  ;Close loop.
        pop ecx                        ;Restore number of DWords to process.
        std                            ;Process in descending addresses
                                       ;(to perform a multiplication the least
                                       ;significant DWords should be taken
                                       ;first).
        sub edi,4                      ;Point to least significant DWord.
        mov esi,Numerator              ;ESI = Multiplicator.
        mov ebx,esi                    ;EBX = Multiplicator / 2 (for rounding)
        shr ebx,1
LoopMultByNumerator:
        mov eax,es:[edi]               ;Get DWord from factor.
        mul esi                        ;Perform multiplication.
        add eax,ebx                    ;Add carry from previous multiplication.
        adc edx,0                      ;Adjust most significant DWord.
        mov ebx,edx                    ;Store carry of multiplication.
        stos dword ptr es:[edi]        ;Store DWord of product.
        eloop LoopMultByNumerator      ;Close loop.
        and eax,eax                    ;The most significant DWord is zero?
        jnz short NumberOfDWordsOfTermOk ;Jump if it is not.
        inc FirstDWord                 ;The first DWord is the next one.
NumberOfDWordsOfTermOk:
        add divisor,2                  ;The divisor 3,5,7,... of the formula
                                       ;must be incremented.
        call PointerSubrPowersOf2      ;Add binary powers to the term.
        mov esi,edi                    ;Index of the most significant DWord
                                       ;in the term.
        mov edi,ecx                    ;EDI = Number of DWords of the number.
        shl edi,3                      ;Point to the beginning of quotient.
        add edi,esi                    ;Point to the most significant DWord
                                       ;of quotient.
        shr esi,2                      ;Convert to number of DWords.
        sub ecx,esi
        jbe short BinaryToDecimalNoRET
        xor edx,edx                    ;Initialize remainder of division.
        mov ebx,divisor                ;Get divisor.
        cld                            ;Process in ascending addresses
                                       ;(to perform a division the most
                                       ;significant DWords should be taken
                                       ;first).
        add esi,NbrDWords              ;
        shl esi,2                      ;Convert to number of bytes.
        push ecx
CicloDivisionPorValorVariable:
        lods dword ptr es:[esi]
        div ebx
        stos dword ptr es:[edi]
        eloop CicloDivisionPorValorVariable
        pop ecx
        mov edi,NbrDWords
        shl edi,2                      ;Convert to number of bytes.
        lea esi,[edi*2+edi-4]          ;Point to the least significant DWord
                                       ;in the term.
        add edi,-4                     ;Point to the least significant DWord
                                       ;in the sum.
        std
        ret
;----------------------------------------------------------------------------
BinaryToDecimalNoRET:
        pop ax                         ;Do not go back to caller routine.
BinaryToDecimal:
        mov al,bufferdos               ;Obtain the integer part (in ASCII).
        mov FormattedNumberBuffer,al   ;Store this value in case of /F.
        mov ax,4400h                   ;Call to IOCTL.
        mov bx,HANDLE_STDOUT           ;Handle STDOUT.
        call dos
        test dl,dl                     ;Bit 7 of DL = 1: Device.
                                       ;            = 0: File.
        jns short FormattedText?       ;Jump if STDOUT is not redirected.
        mov ShowInInt8,0               ;Do not show text in INT 8.
        mov dx,offset TotalTime
        mov cx,2
        mov bx,cx                      ;BX = Handle STDERR.
        mov ah,40h
        call dos
FormattedText?:
        test CommandLineSwitches,Format  ;Should text be formatted?
        jnz short StartConversion        ;Jump if it should.
        mov dx,offset bufferdos
        mov cx,2                       ;Show integer part and point.
        mov bx,handle
        mov ah,40h
        call dos
; Convert from binary to decimal.
StartConversion:
        std
        mov eax,NbrDWords
        lea eax,[eax*4 - 4]
        mov LastDWord,eax
MultiplyByMaxPower:
        mov ecx,NbrDWords
        mov edi,LastDWord
        mov esi,MaxPower                ;Load max power in a DWord.
        xor ebp,ebp                     ;First multiplication carry.
        xor ebx,ebx                     ;Second multiplication carry.
LoopMultByMaxPower:
        mov eax,es:[edi]                ;Obtain DWord from number.
        mul esi                         ;Perform first multiplication.
        add eax,ebp
        adc edx,0
        mov ebp,edx
        mul esi                         ;Perform second multiplication.
        add eax,ebx
        adc edx,0
        mov ebx,edx
        stos dword ptr es:[esi]
        eloop LoopMultByMaxPower
;At this moment EBP has the most significant digits and EBX the next.
EndMultInConv:
        mov eax,LogMaxPower          ;Corresponde a primera multiplicacion.
        add SumLogPower,eax
        jnc short SecondAddLogs
        sub LastDWord,4
        dec NbrDWords
SecondAddLogs:
        add SumLogPower,eax         ;Corresponde a segunda multiplicacion.
        jnc short GetNextDigits
        sub LastDWord,4
        dec NbrDWords
GetNextDigits:
        mov di,offset bufferdos[-1] ;DI = Pointer to start of temporary
                                    ;buffer of digits on screen.
        mov cx,word ptr Exponent    ;Number of digits of number.
        add di,cx
        add di,cx                   ;DI = Pointer to last digit.
        movzx esi,Base              ;ESI = Base.
        call EBX_To_Other_Base
        mov cx,word ptr Exponent    ;Number of digits of number.
        mov ebx,ebp                 ;Convert most significant digits.
        call EBX_To_Other_Base
        mov ecx,digits
        mov eax,Exponent
        add eax,eax
        cmp ecx,eax
        jbe short CantidadDeDigitsAMostrarOK
        mov ecx,eax
CantidadDeDigitsAMostrarOK:
        test CommandLineSwitches,Format  ;Format text?
        jnz short FormatText             ;Jump if so.
        jecxz EndConversion              ;Jump if no decimals were requested.
        mov bx,handle
        mov dx,offset bufferdos
        mov ah,40h
        call dos
        mov eax,Exponent
        shl eax,1
        sub digits,eax
OtherDigits:
        jbe short EndConversion
        jmp MultiplyByMaxPower
FormatText:
        jecxz LastLineFormattedText
        call FormattedDigits
        cmp digits,0
        jnz OtherDigits
LastLineFormattedText:
        call ShowBufferOnScreen
EndConversion:
Final:
        mov dx,OldInt8Offset     ;Set old interrupt handler 8 for
        mov cx,OldInt8Segment    ;real mode again.
        mov ax,0201h
        mov bl,8
        int 31h
        mov dx,offset TotalTime ;CR - LF
        mov cx,2
        mov bx,cx                ;STDERR handle.
        mov ah,40h
        call dos
        mov dx,offset TotalTime
        mov cx,LengthTotalTime
        mov bx,HANDLE_STDERR     ;STDERR handle.
        mov ah,40h
        call dos
        mov ax,0007h             ;Set video buffer descriptor.
        mov bx,Selector
        mov cx,000Bh
        mov dx,8000h
        int 31h
        call getcursor           ;Obtain (in DI) he cursor position.
        mov es,Selector
        call ShowTime            ;Show total execution time.
        mov dx,offset TotalTime  ;CR - LF
        mov cx,2
        mov bx,cx                ;STDERR handle.
        mov ah,40h
        call dos
        mov bx,Selector          ;Free selector requested to DPMI server.
        mov ax,0001h
        int 31h
        mov si,MemoryHandleHigh  ;Free memory block requested to DOS.
        mov di,MemoryHandleLow
        mov ax,0502h
        int 31h
        mov ax,4c00h             ;End of program.
        int 21h
;----------------------------------------------------------------------------
; Subroutine dos: Call DOS from protected mode.
;
dos:
        mov word ptr RealMode.reg_EAX,ax  ;Store registers in structure
        mov word ptr RealMode.reg_EBX,bx  ;used to pass registers to
        mov word ptr RealMode.reg_ECX,cx  ;real mode.
        mov word ptr RealMode.reg_EDX,dx
        mov word ptr RealMode.reg_ESI,si
        mov word ptr RealMode.reg_EDI,di
        mov word ptr RealMode.reg_SS,0
        mov word ptr RealMode.reg_SP,0
        mov ax,0300h                      ;Call Interrupt in real mode.
        mov bx,0021h                      ;DOS = Interrupt 21h.
        xor cx,cx                         ;Number of bytes to pass in stack:
                                          ;zero.
        mov di,offset RealMode            ;Point to real mode structure.
        movzx edi,di
        push ds
        pop es
        int 31h
        mov ah,byte ptr RealMode.reg_Flags
        sahf
        mov es,Selector
        mov ax,word ptr RealMode.reg_EAX
        mov bx,word ptr RealMode.reg_EBX
        mov cx,word ptr RealMode.reg_ECX
        mov dx,word ptr RealMode.reg_EDX
        mov si,word ptr RealMode.reg_ESI
        mov di,word ptr RealMode.reg_EDI
        ret
;----------------------------------------------------------------------------
; Interrupt 8 handler (timer). It must show the number of digits and the
; elapsed time.
;
Int8Handler:
        push ds                        ;Preserve used registers.
        push es
        push esi
        push di
        push bx
        push eax
        push edx
        push cx
        push cs                        ;Load data segment register.
        pop ds                         ;
        cld
        mov di,offset FracSecond
        add dword ptr [di],235903917   ;2^32 / 18,20648193
        jnc short NumberOfDigitsToScreen
        dec di                         ;Point to seconds.
        inc byte ptr [di]              ;Increment seconds.
        cmp byte ptr [di],60           ;Minute finished?
        jnz short TimeActualized       ;Jump if not.
        mov byte ptr [di],0
        dec di                         ;Point to minutes.
        inc byte ptr [di]              ;Increment minutes.
        cmp byte ptr [di],60           ;Hour finished?
        jnz short TimeActualized       ;Jump if not.
        mov byte ptr [di],0
        dec di                         ;Point to hour.
        inc byte ptr [di]              ;Increment hour.
        cmp byte ptr [di],24           ;Day finished?
        jnz short TimeActualized       ;Jump if not.
        mov byte ptr [di],0
        dec di                         ;Point to day.
        inc byte ptr [di]              ;Increment day.
        cmp byte ptr [di],100          ;Have passed 100 days?
        jnz short TimeActualized       ;Jump if they have not.
        mov byte ptr [di],0
TimeActualized:
        cmp ShowInInt8,0               ;Number and time must be shown?
        jz EndInter8                   ;Exit interrupt if not.
        les di,CursorVideo             ;Get cursor location.
        add di,15*2                    ;Point to time.
        call ShowTime                  ;Show elapsed time.
NumberOfDigitsToScreen:
        cmp ShowInInt8,0               ;Number and time must be shown?
        jz EndInter8                   ;Exit interrupt if not.
        les di,CursorVideo             ;Get cursor location.
        mov si,PointerBase
        xor eax,eax
        test CommandLineSwitches,E     ;Is the number e being computed?
        jz short E_IsNotComputed       ;Jump if not.
        mov edx,67108864               ;2^64 / 64
        div dword ptr [si]
        mul FirstDWord
        jmp short AdjustDigits
E_IsNotComputed:
        mov edx,6291456                ;2^64 * log 8 / log (2^32) / 64.
        div dword ptr [si]             ;2^32 * log 8 / log Base
        mul Divisor                    ;EDX = Number of digits / 64.
AdjustDigits:
        shld edx,eax,6                 ;EDX = Number of digits.
        mov bl,0
        mov esi,1000000000
        call Div1Digit
        mov esi,100000000
        call Div1Digit
        mov esi,10000000
        call Div1Digit
        mov esi,1000000
        call Div1Digit
        mov esi,100000
        call Div1Digit
        mov esi,10000
        call Div1Digit
        mov si,1000
        call Div1Digit
        mov si,100
        call Div1Digit
        mov si,10
        call Div1Digit
        add dl,"0"
        mov es:[di],dl
EndInter8:
        pop cx                  ;Restaurar los registros afectados.
        pop edx
        pop eax
        pop bx
        pop di
        pop esi
        pop es
        pop ds
        jmp cs:OldInt8Handler   ;Go to old INT 8 handler.
;---------------------------------------------------------------------------
SetInterruptVector:
        mov dx,offset TextSecondLine
        mov cx,LengthTextSecondLine
        mov bx,HANDLE_STDERR    ;STDERR handle.
        mov ah,40h
        call dos                ;Show text.
        call getcursor          ;Get cursor location.
        mov CursorVideoOffset,di
        mov ax,0200h            ;Get interrupt vector (real mode).
        mov bl,8                ;Interrupt number.
        int 31h
        mov OldInt8Offset,dx    ;Preserve vector.
        mov OldInt8Segment,cx
        mov ax,0201h            ;Skip interrupt vector (real mode).
        mov bl,8                ;Interrupt number.
        mov cx,RealMode.reg_DS
        mov dx,offset Int8Handler
        int 31h
        ret                     ;End of subroutine.
;---------------------------------------------------------------------------
; Subrutina para obtener la posicin del cursor en la pantalla.
; Salida: DI = Posicion en buffer de video que corresponde al cursor.
;
getcursor:mov dx,port6845    ;DX = Video controller control port.
        mov al,0Eh           ;Tell 6845 to read the higher byte (MSB)
        out dx,al            ;of cursor location.
        jmp $+2
        jmp $+2
        inc dx               ;Point to data port of 6845.
        in al,dx             ;Read value.
        mov ah,al            ;Store it in MSB of AX.
        jmp $+2
        jmp $+2
        dec dx               ;DX = Video controller control port.
        mov al,0Fh           ;Tell 6845 to read the lower byte (LSB)
        out dx,al            ;of cursor location.
        jmp $+2
        jmp $+2
        inc dx               ;Point to data port of 6845.
        in al,dx             ;Read value.
        add ax,ax
        mov di,ax            ;Store cursor location in DI.
        ret                  ;End of subroutine.
;----------------------------------------------------------------------------
; Subrutina ShowTime: Show elapsed time.
; Entrada: ES:DI = Video buffer location.
;
ShowTime:
        mov si,offset Days
        mov cx,4
        cld
LoopShowTime:
        lodsb
        aam
        add ax,"00"
        mov es:[di],ah
        mov es:[di+2],al
        mov al,[si-5]
        mov es:[di+4],al
        mov byte ptr es:[di+6]," "
        add di,8
        loop LoopShowTime
        ret
;-----------------------------------------------------------------------------
Div1Digit:
        mov eax,edx
        xor edx,edx
        div esi
        add al,"0"
        cmp al,"0"
        jnz short ShowTheDigit
        test bl,bl
        jz short PointToNextDigit
ShowTheDigit:
        mov bl,1
        stosb
        inc di
PointToNextDigit:
        ret
;-----------------------------------------------------------------------------
; Subrutina FormattedDigits: Halla un digito del numero y, si esta
; habilitado con el switch /F, realiza el formateo de texto.
;
; Entrada: bufferdos: Digits en ASCII.
;          CX = Number of digits a mostrar.
;          DI = Puntero al buffer donde se almacenan los digits en ASCII.
;
FormattedDigits:
        push es                      ;Preserve pointer to binary number.
        push ds
        pop es
        cld
        mov si,offset bufferdos
        mov dx,si
        add dx,word ptr Exponent
        add dx,word ptr Exponent
        mov ah,0                     ;Show most significant zeros as spaces.
        test CommandLineSwitches,ShowNumberOfDigits
        jnz short LoopShowWithoutFormatting
        test CommandLineSwitches,Format
        jnz short ShowFormatted
        mov ah,1                     ;Show most significant zeros as zeros.
LoopShowWithoutFormatting:
        lodsb
        cmp al,"0"
        jnz short SignificantDigit
        test ah,ah
        jnz short SignificantDigit
        cmp si,dx
        jz short ShowDigit
        mov al," "
        jmp short ShowDigit
SignificantDigit:
        mov ah,1
ShowDigit:
        stosb
        loop LoopShowWithoutFormatting
        std
        pop es
        ret
ShowFormatted:
        dec digits
        mov di,PositionDigitInBuffer
        mov bx,FormattingInformation    ;BH = Group number (0-9).
                                        ;BL = Digit inside group (1-5).
        inc bx                          ;Indicate next digit inside group.
        cmp bl,6                        ;End of group?
        jnz short PutFormattedDigit     ;Jump if not.
        mov bl,1                        ;Indicate start of group.
        inc bh                          ;Indicate next group.
        cmp bh,10                       ;End of line?
        jnz short PutSpace          ;Go to show space if not.
        push si
        call ShowBufferOnScreen
        pop si
        mov di,offset FormattedNumberBuffer
                                ;Clear formatted digit buffer.
        push cx                 ;Preserve shown digits counter.
        cld
        mov cx,75
        mov al," "
        rep stosb
        pop cx                  ;Restore shown digits counter.
        mov bx,HANDLE_STDOUT
        mov di,offset FormattedNumberBuffer + 1
                                ;Point to second byte of buffer.
PutSpace:
        inc di                  ;Point to next character.
PutFormattedDigit:
        movsb                   ;Copy digit in formatted digit buffer.
        mov PositionDigitInBuffer,di
        mov FormattingInformation,bx
        inc DigitsShown         ;Indicate another digit was shown.
        loop ShowFormatted
        std
        pop es
        ret                     ;End of subroutine.
;---------------------------------------------------------------------------
ShowBufferOnScreen:
        push cx
        or CommandLineSwitches,ShowNumberOfDigits
        push esi
        mov ebx,DigitsShown
        mov di,offset FormattedNumberBuffer[74]
                                ;DI = Pointer to end of number shown at
                                ;     extreme right buffer.
        mov esi,10              ;ESI = Base.
        mov cx,13               ;CX = Number of digits to convert.
        call EBX_To_Other_Base
        mov cx,12
LoopDiscardLeftZeros:
        inc di
        cmp byte ptr [di],"0"
        jnz short DiscardedZeros
        mov byte ptr [di]," "
        loop LoopDiscardLeftZeros
DiscardedZeros:
        pop esi
        and CommandLineSwitches,not ShowNumberOfDigits
        mov bx,handle
        mov dx,offset FormattedNumberBuffer
        mov cx,79
        push es
        mov ah,40h
        call dos
        pop es
        pop cx
        ret
;------------------------------------------------------------------------------
; Conversion from binary (EBX = 32 bits) to another base.
;
; Entrada: EBX = Doble palabra a convertir.
;          ESI = Base.
;          CX = Number of digits a convertir.
; Salida:  bufferdos: Numero en ASCII.
;
EBX_To_Other_Base:
LoopBinToDec:
        mov eax,ebx             ;Load dividend.
        xor edx,edx             ;Extend dividend.
        div esi                 ;Perform division.
        mov ebx,eax             ;Save quotient, it will be the dividend.
        add dl,"0"              ;Convert to ASCII.
        cmp dl,"9"
        jbe short PutDigit
        add dl,"A" - ("9" + 1)
PutDigit:
        mov [di],dl             ;Save remainder in temporary buffer
                                ;(in this way the least significant digit
                                ;or the digit "at the right" is found).
        dec di                  ;Point to previous digit.
        loop LoopBinToDec    ;Jump if buffer is not over.
        ret                     ;End of subroutine.
;---------------------------------------------------------------------------
; Subroutine switch: Process command line switches initializing the
; corresponding flags.
;
; On entry: BX = Pointer to command line buffer.
;
switch: cmp byte ptr [bx],"/"            ;Switch character found?
        jz short SwitchFound        ;Jump if so.
        cmp byte ptr [bx],ASCII_CR       ;Command line end?
        jz short EndSwitch               ;No switches if so.
        inc bx                           ;Point to next char in cmd line.
        cmp byte ptr [bx-1]," "          ;Space or tab character?
        jbe switch                       ;Skip it if so.
        lea dx,[bx-1]                    ;DX = Pointer to file name.
LookForEndOfFileName:
        inc bx                           ;Point to next character.
        cmp byte ptr [bx-1]," "          ;End of file name?
        ja LookForEndOfFileName          ;Continue looking for end of file 
                                         ;name if not.
        mov [bx-1],bh                    ;Convert name to ASCIIZ format.
        mov ah,3Ch                       ;Try to create file.
        xor cx,cx
        call dos
        jc short JmpToError              ;Jump if file could not be created.
        mov handle,ax                    ;Store file handle.
EndSwitch:
        ret                              ;End of subroutine.
JmpToError:
        jmp syntax_error                 ;Go to error routine.
SwitchFound:
        mov ax,[bx+1]                    ;Get switch type.
        and ax,0DFDFh                    ;Convert to capital letters.
        cmp al,"E"                       ;Do the user want to compute e?
        jnz short IsLN2?                 ;Jump if he does not.
        or CommandLineSwitches,E         ;Indicate to compute e.
        inc bx                           ;Skip switch.
        inc bx
        jmp switch                       ;Look for more switches.
IsLN2?:
        cmp ax,"NL"                      ;First part of LN 2?
        jnz short IsFormat?              ;Jump if not.
        cmp byte ptr [bx+3],"2"          ;Second part of LN 2?
        jnz short IsFormat?              ;Jump if not.
        or CommandLineSwitches,LN2       ;Indicate to compute ln 2.
        add bx,4                         ;Skip switch.
        jmp switch                       ;Look for more switches.
IsFormat?:
        cmp al,"F"                       ;Text formatted switch?
        jnz short IsBase?                ;Jump if not.
        or CommandLineSwitches,Format    ;Indicate text is to be formatted.
        add bx,2                         ;Skip switch.
        jmp switch                       ;Look for more switches.
IsBase?:
        mov ax,[bx+1]
        and al,0DFh                      ;Convert first char to caps.
        cmp ax,":B"                      ;Is base?
        jnz JmpToError                   ;Go to error routine on
                                         ;invalid switch.
        mov cl,0                         ;Initialize base.
        add bx,3                         ;Point to digits.
LoopBase:
        mov ch,[bx]                      ;Obtener el byte de la linea.
        sub ch,"0"                       ;Convert from ASCII to binary.
        cmp ch,9                         ;Is a digit?
        ja short EndBase                 ;End searching digits if not.
        mov al,10                        ;Multiply base by 10.
        mul cl
        jc JmpToError                    ;Error if base is greater than 256.
        add al,ch                        ;Get new base.
        cmp al,36                        ;Is base greater than 36?
        ja JmpToError                    ;Error if it is.
        mov cl,al                        ;Preserve value.
        inc bx                           ;Point to next byte.
        jmp LoopBase                     ;Get more digits.
EndBase:
        cmp cl,2                         ;Is base less than 2?
        jb JmpToError                    ;Error if it is.
        mov Base,cl                      ;Store base.
        jmp switch
;-----------------------------------------------------------------------------
EndCode:
PIDPMIE ends
        end begin
